----------------------------------------------------------------------------------------------------
---- ,    CRC.                                                     ----
----   CompressionMethod, Compressor, UserCompressor -   .        ----
----      ,    .               ----
----------------------------------------------------------------------------------------------------
module Compression where

import Control.Concurrent
import Control.Monad
import Data.Bits
import Data.Char
import Data.IORef
import Data.List
import Data.Maybe
import Data.Word
import Foreign.C.String
import Foreign.C.Types
import Foreign.Marshal.Alloc
import Foreign.Marshal.Pool
import Foreign.Ptr
import System.IO.Unsafe

import CompressionLib hiding (b,kb,mb,gb)
import Errors
import Utils
import Files
import qualified ByteStream


-- |      
type CompressionMethod  =  CompressionLib.Method

--  "",  ,    CompressionLib
aSTORING              = "storing"
aFAKE_COMPRESSION     = "fake"
aCRC_ONLY_COMPRESSION = "crc"

-- | ()  
isFakeMethod     method  =  method==aFAKE_COMPRESSION  ||  method==aCRC_ONLY_COMPRESSION
-- |LZP  
isLZP_Method     method  =  method_name method == "lzp"
-- |Tornado  
isTornado_Method method  =  method_name method == "tor"
-- |DICT  
isDICT_Method    method  =  method_name method == "dict"
-- |TTA  
isTTA_Method     method  =  method_name method == "tta"
-- |MM  
isMM_Method      method  =  method_name method == "mm"
-- |JPG  
isJPG_Method     method  =  method_name method == "jpg"
-- |GRZip  
isGRZIP_Method   method  =  method_name method == "grzip"
-- |    (>10 mb/s  1 )
isVeryFastMethod         =  compressionIs "VeryFast?"
-- |  
isFastDecMethod          =  not . any_function [=="ppmd", =="ppmm", =="pmm", compressionIs "external?"] . method_name


-- |  ,    
type Compressor = [CompressionMethod]

-- | "storing" (-m0)
aNO_COMPRESSION = [aSTORING] :: Compressor

-- | -  ,          - 
isFakeCompressor (method:xs)  =  isFakeMethod method  &&  null xs

-- | - LZP ,          - LZP
isLZP_Compressor (method:xs)  =  isLZP_Method method  &&  null xs

-- | -   ,     ,    
isVeryFastCompressor (method:xs)  =  isVeryFastMethod method  &&  null xs

-- | -  ,       
isFastDecompressor = all isFastDecMethod


-- |       .
--       , 
--   (    ,     )
type UserCompressor = [(String,Compressor)]  --    "$text->m3t, $exe->m3x, $compressed->m0"

getCompressors :: UserCompressor -> [Compressor]
getCompressors = map snd

getMainCompressor :: UserCompressor -> Compressor
getMainCompressor = snd.head

-- | -  Storing,       aNO_COMPRESSION    
isStoring ((_,compressor):xs)  =  compressor==aNO_COMPRESSION  &&  null xs

-- | - fake compression,           
isFakeCompression ((_,compressor):xs)  =  isFakeCompressor compressor  &&  null xs

-- | - LZP compression,      LZP     
isLZP_Compression ((_,compressor):xs)  =  isLZP_Compressor compressor  &&  null xs

-- |   ,            
isVeryFastCompression = all (isVeryFastCompressor.snd)

-- |  ,           
isFastDecompression = all (isFastDecompressor.snd)

-- | ,      `ftype`.
--           -  
--  ,     
findCompressor ftype list  =  lookup ftype list  `defaultVal`  snd (head list)

-- |         
instance ByteStream.BufferData Compressor where
  write buf x  =  ByteStream.write buf (join_compressor x)
  read  buf    =  ByteStream.read  buf  >>==  split_compressor


-------------------------------------------------------------------------------------------------------------
-- Encode/decode compression method for parsing options/printing info about selected compression method -----
-------------------------------------------------------------------------------------------------------------

-- |Parse command-line option that represents compression method.
--        ,     
-- "  ->  ".         
decode_method configuredMethodSubsts str =
    str                       -- "3/$obj=2b/$iso=ecm+3b"
    .$ subst list             -- "3b/3t/$obj=2b/$iso=ecm+3b"
    .$ split_to_methods       -- [("","exe+3b"), ("$obj","3b"), ("$text","3t"), ("$obj","2b"), ("$iso","ecm+3b")]
    .$ keepOnlyLastOn fst     -- [("","exe+3b"), ("$text","3t"), ("$obj","2b"), ("$iso","ecm+3b")]
    .$ filter (not.null.snd)  -- "-m$bmp="        $bmp
    .$ mapSnds (subst2 list)  -- [("",["exe","lzma"]), ("$text",["ppmd"]), ("$obj",["lzma"]), ("$iso",["ecm","lzma"])]

    where list = prepareSubsts (concatMap reorder [configuredMethodSubsts, builtinMethodSubsts])   --   ,  ,    
          reorder list = a++b  where (a,b) = partition (notElem '#') list                          --   :  ,   #,   # ( ,   )

--       (     )
subst list method  =  joinWith "/" (main_methods:group_methods++user_methods)
  where --    -m3/$obj=2b      ,  
        main:user_methods = split '/' method
        --    ,  3x = 3xb/3xt
        main_methods = case (lookup main list) of
            Just x  -> subst list x   --    
            Nothing -> main           --   
        --          ,  3x$iso = ecm+exe+3xb
        group_methods = list .$ keepOnlyFirstOn fst                      --    (      ,    )
                             .$ mapMaybe (startFrom main . join2 "=")    --   ,   3x,   3x
                             .$ filter (("$"==).take 1)                  --    -    $

--       (-     )
subst2 list  =  concatMap f . split_compressor
    where f method = let (head,params)  =  break (==':') method
                     in case (lookup head list) of
                          Just new_head -> subst2 list (new_head++params)
                          Nothing       -> [decode_one_method method]

-- |    
decode_one_method method | isFakeMethod method = method
                         | otherwise           = canonizeCompressionMethod method

--   ,       ,
--    ( ,  )
split_to_methods method = case (split '/' method) of
    [_]                 ->  [("",method)]   --      
    x : xs@ (('$':_):_) ->  ("",x) : map (split2 '=') xs   -- m1/$type=m2...
    b : t : xs          ->  [("","exe+"++b), ("$obj",b), ("$text",t)] ++ map (split2 '=') xs   -- m1/m2/$type=m3...

--       lookup
prepareSubsts x = x
    --   ,   
    .$ map (filter (not.isSpace) . fst . split2 ';') .$ filter (not.null)
    --      #  9 ,  #    1  9
    .$ concatMap (\s -> map (\d->replace '#' d s) ['1'..'9'])
    --     "a=b"    lookup
    .$ map (split2 '=')

--      ,    arc.ini
builtinMethodSubsts = [
      ";High-level method definitions"
    , "x  = 9            ;highest compression mode using only internal algorithms"
    , "ax = 9p           ;highest compression mode involving external compressors"
    , "0  = storing"
    , "1  = 1b  / $exe=exe+1b"
    , "1x = 1"
    , "2  = 2b  / $exe=exe+2b  / $text=2t"
    , "2x = 2xb / $exe=exe+2xb / $text=2xt"
    , "#  = #b/#t        ;equivalent to the following:  exe+#b / $obj=#b / $text=#t"
    , "#r = #br/#tr      ; exe+#b / $obj=#b / $text=#tr"
    , "#x = #bx/#tx      ; exe+#xb/ $obj=#xb/ $text=#xt"
    , ""
    , ";Text files compression with slow decompression"
    , "1t  = 1b"
    , "2t  = grzip:m4:8m:32:h15"
    , "3tr = grzip:m1:8m:32:h15"
    , "3t  = dict:p: 32m:80% + lzp: 32m:055:d1m:s16:h18:87% + ppmd:7:48m"    --  -m3t   ppmd  grzip(bwt)
    , "4t  = dict:p: 64m:80% + lzp: 64m:065:d1m:s16:h20:90% + ppmd:8:96m"
    , "5t  = dict:p: 64m:80% + lzp: 64m:105:d1m:s32:h22:92% + ppmd:12:192m"
    , "6t  = dict:p:128m:80% + lzp:128m:145:d1m:s32:h23:92% + ppmd:16:384m"
    , "7t  = dict:p:128m:80% + lzp:128m:185:d1m:s32:h24:92% + ppmd:20:768m"
    , "8t  = dict:p:128m:80% + lzp:128m:225:d1m:s32:h24:92% + ppmd:24:1536m"
    , "9t  = 8t"
    , "4tr = 4t:12:r"     --  m4t   "-r1"  ppmd
    , "5tr = 5t:16:r"     --  m5t   "-r1"  ppmd
    , "6tr = 6t:20:r"     --  m6t   "-r1"  ppmd
    , "#tr = #t"
    , ""
    , ";Binary files compression with slow and/or memory-expensive decompression"
    , "1b  = 1xb"
    , "2b  = 2xb"
    , "2br = rep:  16m + lzma:2m:fastest"
    , "3b  = rep:  32m + 3xb"
    , "4b  = rep:  64m + 4xb"
    , "5b  = rep: 128m + 5xb"
    , "6b  = rep: 256m + 6xb"
    , "7b  = rep: 512m + 7xb"
    , "8b  = rep:1024m + 8xb"
    , "9b  = 8b"
    , "#br = #b"
    , ""
    , ";Text files compression with fast decompression"
    , "1xt = 1xb"
    , "2xt = 2xb"
    , "3xt = dict:  32m:80% + 3xb"
    , "4xt = dict:  64m:75% + 4xb"
    , "5xt = dict: 128m:75% + 5xb"
    , "6xt = dict: 128m:75% + 6xb"
    , "7xt = dict: 128m:75% + 7xb"
    , "8xt = dict: 128m:75% + 8xb"
    , "9xt = dict: 128m:75% + 9xb"
    , ""
    , ";Binary files compression with fast decompression"
    , "1xb = tor:3"
    , "2xb = tor"
    , "3xb = delta + lzma:  4m:fast:mc8"
    , "4xb = delta + lzma:  8m:normal:mc16"
    , "5xb = delta + lzma: 16m:max"
    , "6xb = delta + lzma: 32m:max"
    , "7xb = delta + lzma: 64m:max"
    , "8xb = delta + lzma:128m:max"
    , "9xb = 8xb"
    , ""
    , ";Synonyms"
    , "bcj = exe"
    , "#rt = #tr"
    , "#rb = #br"
    , "#bx = #xb"
    , "#tx = #xt"
    , "x#  = #x"    --    "-mx7"    7-zip
    , ""
    , ";Compression modes involving external PPMONSTR.EXE"
    , "#p  = #pb/#pt"
    , "5pt = dict:p: 64m:80% + lzp: 64m:32:h22:85% + pmm: 8:160m:r0"
    , "6pt = dict:p: 64m:80% + lzp: 64m:64:h22:85% + pmm:16:384m:r1"
    , "7pt = dict:p:128m:80% + lzp:128m:64:h23:85% + pmm:20:768m:r1"
    , "8pt = dict:p:128m:80% + lzp:128m:64:h23:85% + pmm:24:1536m:r1"
    , "9pt = 8pt"
    , "#pt = #t"
    , "#pb = #b"
    , ""
    , "#q  = #qb/#qt"
    , "5qt = dict:p:64m:80% + lzp:64m:64:d1m:24:h22:85% + pmm:10:160m:r1"
    , "5qb = rep: 128m                                  + pmm:16:160m:r1"
    , "6qb = rep: 256m                                  + pmm:20:384m:r1"
    , "7qb = rep: 512m                                  + pmm:22:768m:r1"
    , "8qb = rep:1024m                                  + pmm:24:1536m:r1"
    , "9qb = 8qb"
    , "#qt = #pt"
    , "#qb = #pb"
    , ""
    , ";Sound wave files are best compressed with TTA algorithm"
    , "wav     = tta      ;best compression"
    , "wavfast = tta:m1   ;faster compression and decompression"
    , "1$wav  = wavfast"
    , "2$wav  = wavfast"
    , "#$wav  = wav"
    , "#x$wav = wavfast"
    , "#p$wav = wav"
    , ""
    , ";Bitmap graphics files are best compressed with GRZip algorithm"
    , "bmp        = mm    + grzip:m1:l:a  ;best compression"
    , "bmpfast    = mm    + grzip:m4:l:a  ;faster compression"
    , "bmpfastest = mm:d1 + tor:2         ;fastest one"
    , "1$bmp  = bmpfastest"
    , "2$bmp  = bmpfastest"
    , "3$bmp  = bmpfast"
    , "#$bmp  = bmp"
    , "1x$bmp = bmpfastest"
    , "2x$bmp = bmpfastest"
    , "#x$bmp = mm+#xb"
    , "#p$bmp = bmp"
    ]


-- |Human-readable description of compression method
encode_method uc  =  joinWith ", " (map encode_one_method uc)
encode_one_method (group,compressor)  =  between group " => " (join_compressor compressor)
join_compressor  =  joinWith "+"

-- |Opposite to join_compressor (used to read compression method from archive file)
split_compressor  =  split '+'

-- |      process
process_algorithms process compressor = do
    return (split_compressor compressor)
       >>=  mapM process
       >>== join_compressor

-- |       
split_method = split ':'

-- |  
method_name = head . split_method

-- |,      
showMem 0      = "0b"
showMem mem    = showM [(gb,"gb"),(mb,"mb"),(kb,"kb"),(b,"b"),error"showMem"] mem

showMemory 0   = "0 bytes"
showMemory mem = showM [(gb," gbytes"),(mb," mbytes"),(kb," kbytes"),(b," bytes"),error"showMemory"] mem

showM xs@( (val,str) : ~(nextval,_) : _) mem =
  if mem `mod` val==0 || mem `div` nextval>=1000
    then show((mem+val`div` 2) `div` val)++str
    else showM (tail xs) mem

-- |    ,    
roundMem mem | mem>=3999*kb = mem `roundUp` mb
             | otherwise    = mem `roundUp` kb

{-# NOINLINE builtinMethodSubsts #-}
{-# NOINLINE decode_method #-}
{-# NOINLINE showMem #-}


----------------------------------------------------------------------------------------------------
-----                                                              -----
----------------------------------------------------------------------------------------------------

-- |  /    / 
--    ,   UserCompressor
compressorSetDictionary               = map . mapSnd . mapLast . freearcSetDictionary
compressorLimitCompressionMem         = map . mapSnd . map . freearcLimitCompressionMem
compressorLimitDecompressionMem       = map . mapSnd . map . freearcLimitDecompressionMem
compressorLimitDictionary             = map . mapSnd . compressionLimitDictionary
compressorLimitCompressionMemoryUsage = map . mapSnd . compressionLimitMemoryUsage freearcGetCompressionMem

-- |    /     compressor'
compressorGetCompressionMem      = maximum . map (calcMem freearcGetCompressionMem . snd)
compressorGetDecompressionMem    = maximum . map (calcMem freearcGetDecompressionMem . snd)
compressorGetBlockSize           = maximum . map (maximum . map freearcGetBlockSize . snd)


-- |    ,      ,
--      ( precomp).   
--  ,        :)
compressionLimitDictionary mem (x:xs) =  new_x : (not(compressionIs "external?" new_x)  &&&  compressionLimitDictionary mem) xs
                                             where new_x = freearcLimitDictionary mem x
compressionLimitDictionary mem []     =  []

-- |  tempfile   ,    "",
--   memory_limit+5% (  ""      ).
--    dict/dict+lzp     (blocksize*2  , blocksize/2  ),
--  external compressors   
compressionLimitMemoryUsage getMem memory_limit = go (0::Double) ""
  where go _   _    []      =  []
        go mem prev (x:xs) | compressionIs "external?" x  =  x: go 0            x xs
                           | mem+newMem < memlimit*1.05   =  x: go (mem+newMem) x xs
                           | otherwise                    =  "tempfile":x: go newMem x xs

           where newMem | mem==0 && isDICT_Method x             =  realToFrac (freearcGetBlockSize x) / 2
                        | isDICT_Method prev && isLZP_Method x  =  0
                        | otherwise                             =  realToFrac$ getMem x
                 memlimit = realToFrac memory_limit

-- |          
--    compressionIs "external?"    dict/dict+lzp
calcMem getMem  = maximum . map getMemSum . splitOn (compressionIs "external?")
  where getMemSum (x:y:xs) | isDICT_Method x && isLZP_Method y  =  max (getMem x) ((freearcGetBlockSize x `div` 2) + getMemSum xs)
        getMemSum (x:xs)   | isDICT_Method x                    =  max (getMem x) ((freearcGetBlockSize x `div` 2) + getMemSum xs)
        getMemSum (x:xs)                                        =  getMem x + getMemSum xs
        getMemSum []                                            =  0

-- |    "tempfile"    
compressionDeleteTempCompressors = filter (/="tempfile")


-- |//  /      /
freearcSetDictionary          =  liftSetter . CompressionLib.setDictionary
freearcLimitDictionary        =  liftSetter . CompressionLib.limitDictionary
freearcLimitCompressionMem    =  liftSetter . CompressionLib.limitCompressionMem
freearcLimitDecompressionMem  =  liftSetter . CompressionLib.limitDecompressionMem
freearcGetCompressionMem      =  liftGetter   CompressionLib.getCompressionMem
freearcGetDecompressionMem    =  liftGetter   CompressionLib.getDecompressionMem
freearcGetBlockSize           =  liftGetter   CompressionLib.getBlockSize

-- |   CompressionLib,  Method,  ,  CompressionMethod
liftSetter action  method | aSTORING ==  method   =  method
liftSetter action  method | isFakeMethod method   =  method
liftSetter action  method                         =  action method

-- |   CompressionLib,  Method,  ,  CompressionMethod
liftGetter action  method | aSTORING ==  method   =  0
liftGetter action  method | isFakeMethod method   =  0
liftGetter action  method                         =  action method


----------------------------------------------------------------------------------------------------
----- (De)compression of data stream                                                           -----
----------------------------------------------------------------------------------------------------

{-
compress   comprMethod reader writer      -  
decompress comprMethod reader writer      -  

  comprMethod :: CompressionMethod -  
  reader buf size -      `buf`  `size`
                     0   -  
                               <0  -  () (     )
                               >0  - -  
  writer buf size -   
                     <0  -  () (     )
                               >=0 -  
                            "",  
                      ()         
       ()
-}

-- |     
freearcCompress   num method | aSTORING ==  method =  copy_data
freearcCompress   num method | isFakeMethod method =  eat_data
freearcCompress   num method                       =  checkingCtrlBreak num (CompressionLib.compress method)

-- |     
freearcDecompress num method | aSTORING ==  method =  copy_data
freearcDecompress num method | isFakeMethod method =  undecompressible   --       
freearcDecompress num method                       =  checkingCtrlBreak num (CompressionLib.decompress method)

-- | Haskell' ,   ,   
-- ,    /  
checkingCtrlBreak num action read write callback = do
  let checked_callback what buf size auxdata = do
        aProgramTerminated <- val programTerminated
        if aProgramTerminated
          then forever doNothing0
          else case what of
                 "read"  -> read  buf size
                 "write" -> write buf size
                 _       -> callback what buf size
  --
  res <- checked_callback "read" nullPtr 0 undefined
  if res<0  then return res
            else action (checked_callback)

-- |    (-m0)
copy_data reader writer callback = do
  allocaBytes aHUGE_BUFFER_SIZE $ \buf -> do  --  `alloca`,       
    let go ptr = do
          len <- reader ptr ((buf+:aHUGE_BUFFER_SIZE)-:ptr)
          if (len>0)
            then do let newptr = ptr+:len
                    if newptr < (buf +: aHUGE_BUFFER_SIZE)
                       then go newptr
                       else do result <- writer buf (newptr-:buf)
                               if (result>=0)
                                 then go buf
                                 else return (result)  --   ,   /   
            else do if (len==0 && ptr>buf)
                      then do result <- writer buf (ptr-:buf)
                              return (if result>0 then 0 else result)
                      else return len  --  0,   ,   ,   /   
    go buf --  

-- | ,   ,  CRC     ;)
eat_data reader writer callback = do
  allocaBytes aBUFFER_SIZE $ \buf -> do  --  `alloca`,       
    let go = do
          len <- reader buf aBUFFER_SIZE
          if (len>0)
            then go
            else return len   --  0,   ,   ,   /   
    go  --  

undecompressible reader writer callback = do
  return aFREEARC_ERRCODE_GENERAL   --   ,    (FAKE/CRC_ONLY)   

{-# NOINLINE checkingCtrlBreak               #-}
{-# NOINLINE copy_data                       #-}
{-# NOINLINE eat_data                        #-}
{-# NOINLINE freearcCompress                 #-}
{-# NOINLINE freearcDecompress               #-}
{-# NOINLINE compressorSetDictionary         #-}
{-# NOINLINE compressorLimitCompressionMem   #-}
{-# NOINLINE compressorLimitDecompressionMem #-}
{-# NOINLINE compressorGetCompressionMem     #-}
{-# NOINLINE compressorGetDecompressionMem   #-}
{-# NOINLINE compressorGetBlockSize          #-}


----------------------------------------------------------------------------------------------------
----- CRC calculation ------------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------

-- |CRC 
type CRC  = CUInt
aINIT_CRC = 0xffffffff  :: CRC
updateCRC addr len  =  c_UpdateCRC addr (i len)
finishCRC = xor aINIT_CRC

-- | CRC   
calcCRC addr len  =  updateCRC addr len aINIT_CRC  >>==  finishCRC

-- |Fast C routine for CRC-32 calculation
foreign import ccall unsafe "Environment.h UpdateCRC"
   c_UpdateCRC :: Ptr CChar -> CUInt -> CRC -> IO CRC

